
#include "cudaKernels.h"
#include <cstdio>
#include <sdkHelper.h>  // helper for shared functions common to CUDA SDK samples
#include "cuPrintf.cu"

// parameters for nnls
#include "headers.h"

// for macro
#include "enNNLS.h"


////////////////////////////////////////////////////////////////////////////////
// CUDA kernels

__global__ void cuPrintfExample()
{
	int tid;
	tid = blockIdx.x * blockDim.x + threadIdx.x;
	cuPrintf("%d\n", tid);
	printf("Is this also ok??\n"); // this also works. (without index) sm_20 or more
}


// Matrix multiplication on the device: C = A * B (32x40) * (40x1)
__global__ void matrixMul_NNLS(float* C, float* A, float* B)
{
	// Block index
	unsigned short sysID = blockIdx.x;

	// Thread index
	unsigned short tID = threadIdx.x;

	__shared__ float	sA[Nnnls*Mnnls]; // stores A's elements
	__shared__ float	sB[Mnnls];

	// load A and B
	for (int i=0 ; i < (Nnnls*Mnnls) ; i++) {
		sA[i] = A[sysID*(Nnnls*Mnnls) + i];
	}
	__syncthreads();
	for (int i=0 ; i < Mnnls ; i++) {
		sB[i] = B[sysID*Mnnls + i];
	}
	__syncthreads();

	float Cvalue = 0;
	for (int e=0 ; e < Mnnls ; e++) {
		Cvalue += sA[tID*Mnnls + e]*sB[e];
	}
	__syncthreads();

	C[sysID*Nnnls + tID] = Cvalue;
}


__global__ void
calcPrior_kernel(float* d_A, float* d_B, int rowN, int columnN, int sliceN, float NL_control)
{
	int ind;
	int volumeSize = rowN*columnN*sliceN;

	// Thread index
	int tx = threadIdx.x; // 21
	int ty = threadIdx.y; // 21
	int bx = blockIdx.x;
	int by = blockIdx.y;
	int bz = blockIdx.z;
	int kernelSize = 7*7;
	float T2_kernel_c[40*49]; // 7x7 kernel for a voxel being processed
	float T2_kernel_n[40*49]; // 7x7 kernel for a neighbor voxel

	int xcord, ycord;
	int xcord_c, ycord_c;
	int kernelCount;
	float SKLdiv; // SKL divergence
	float kernel_weight;
	float h = NL_control; // controls a degree of filtering

	// get kernel for a voxel being processed
	kernelCount = 0;
	for (int k=-3; k<4; k++) {
		for (int l=-3; l<4; l++) {
			xcord = bx + l;
			ycord = by + k;
			if (xcord < 0) xcord = 0;
			if (xcord >= rowN) xcord = rowN-1;
			if (ycord < 0) ycord = 0;
			if (ycord >= columnN) ycord = columnN-1;
			ind = xcord + rowN*ycord + (rowN*columnN)*bz;
			for (int t=0 ; t<40; ++t) T2_kernel_c[t + 40*kernelCount] = d_A[ind + t*volumeSize];
			kernelCount++;
		}
	}

	// search window size 21x21
	xcord_c = bx + (tx-10);
	ycord_c = by + (ty-10);
	if (xcord_c < 0) xcord_c = 0;
	if (xcord_c >= rowN) xcord_c = rowN-1;
	if (ycord_c < 0) ycord_c = 0;
	if (ycord_c >= columnN) ycord_c = columnN-1;

	// get kernel for a neighbor voxel
	kernelCount = 0;
	for (int k=-3; k<4; k++) {
		for (int l=-3; l<4; l++) {
			xcord = xcord_c + l;
			ycord = ycord_c + k;
			if (xcord < 0) xcord = 0;
			if (xcord >= rowN) xcord = rowN-1;
			if (ycord < 0) ycord = 0;
			if (ycord >= columnN) ycord = columnN-1;
			ind = xcord + rowN*ycord + (rowN*columnN)*bz;
			for (int t=0 ; t<40; ++t) T2_kernel_n[t + 40*kernelCount] = d_A[ind + t*volumeSize];
			kernelCount++;
		}
	}

	// calculate the SKL div. between kernels
	SKLdiv = 0;
	float SKL_tmp[40];
	float tmp_c[40];
	float tmp_n[40];
	float tmp;
	for (kernelCount=0; kernelCount<kernelSize; kernelCount++) {
		// normalize the kernel
		for (int t=0 ; t<40; ++t) {
			T2_kernel_c[t + 40*kernelCount] += 0.000001; // to avoid inf for SKL div.
			tmp_c[t] = T2_kernel_c[t + 40*kernelCount];
			T2_kernel_n[t + 40*kernelCount] += 0.000001; // to avoid inf for SKL div.
			tmp_n[t] = T2_kernel_c[t + 40*kernelCount];
		}

		tmp = 0;
		for (int t=0; t<40; ++t) tmp += tmp_c[t];
		for (int t=0 ; t<40; ++t) T2_kernel_c[t + 40*kernelCount] /= tmp;
		tmp = 0;
		for (int t=0; t<40; ++t) tmp += tmp_n[t];
		for (int t=0 ; t<40; ++t) T2_kernel_n[t + 40*kernelCount] /= tmp;

		// compute SKL div.
		for (int t=0 ; t<40; ++t) {
			SKL_tmp[t] = T2_kernel_c[t + 40*kernelCount]*
					log10(T2_kernel_c[t + 40*kernelCount]/T2_kernel_n[t + 40*kernelCount]);
			SKL_tmp[t] += T2_kernel_n[t + 40*kernelCount]*
					log10(T2_kernel_n[t + 40*kernelCount]/T2_kernel_c[t + 40*kernelCount]);
		}

		tmp = 0;
		for (int t=0; t<40; ++t) tmp += SKL_tmp[t];

		SKLdiv += tmp;
	}
	SKLdiv /= kernelSize;

	// calculate kernel weight
	kernel_weight = (SKLdiv*SKLdiv)/(h*h);
	kernel_weight = exp(-kernel_weight);

	__shared__ float Ks[21][21];
	Ks[tx][ty] = kernel_weight;
	__syncthreads();

	// compute prior T2 dist.
	float T2[40]={0};
	float normC=0;
	for (int m=-10 ; m<11 ; m++) {
		for (int n=-10 ; n<11 ; n++) {
			xcord_c = bx + m;
			ycord_c = by + n;
			if (xcord_c < 0) xcord_c = 0;
			if (xcord_c >= rowN) xcord_c = rowN-1;
			if (ycord_c < 0) ycord_c = 0;
			if (ycord_c >= columnN) ycord_c = columnN-1;
			ind = xcord_c + rowN*ycord_c + (rowN*columnN)*bz;
			for (int t=0 ; t<40; ++t) T2[t] += Ks[m+10][n+10]*d_A[ind + t*volumeSize];
			normC += Ks[m+10][n+10];
		}
	}
	for (int t=0 ; t<40; ++t) T2[t] /= normC;
	ind = bx + rowN*by + (rowN*columnN)*bz;
	for (int t=0 ; t<40; ++t) d_B[ind + t*volumeSize] = T2[t];
}
////////////////////////////////////////////////////////////////////////////////




////////////////////////////////////////////////////////////////////////////////
// These are CUDA Helper functions

// This will output the proper CUDA error strings in the event that a CUDA host call returns an error
#define checkCudaErrors(err)           __checkCudaErrors (err, __FILE__, __LINE__)

inline void __checkCudaErrors( cudaError err, const char *file, const int line )
{
	if(cudaSuccess != err) {
		fprintf(stderr, "%s(%i) : CUDA Runtime API error %d: %s.\n",
				file, line, (int)err, cudaGetErrorString( err ) );
		exit(-1);
	}
}

// This will output the proper error string when calling cudaGetLastError
#define getLastCudaError(msg)      __getLastCudaError (msg, __FILE__, __LINE__)

inline void __getLastCudaError( const char *errorMessage, const char *file, const int line )
{
	cudaError_t err = cudaGetLastError();
	if( cudaSuccess != err) {
		fprintf(stderr, "%s(%i) : getLastCudaError() CUDA error : %s : (%d) %s.\n",
				file, line, errorMessage, (int)err, cudaGetErrorString( err ) );
		exit(-1);
	}
}

// General GPU Device CUDA Initialization
int gpuDeviceInit(int devID)
{
	int deviceCount;
	checkCudaErrors(cudaGetDeviceCount(&deviceCount));
	if (deviceCount == 0) {
		fprintf(stderr, "gpuDeviceInit() CUDA error: no devices supporting CUDA.\n");
		exit(-1);
	}
	if (devID < 0)
		devID = 0;
	if (devID > deviceCount-1) {
		fprintf(stderr, "\n");
		fprintf(stderr, ">> %d CUDA capable GPU device(s) detected. <<\n", deviceCount);
		fprintf(stderr, ">> gpuDeviceInit (-device=%d) is not a valid GPU device. <<\n", devID);
		fprintf(stderr, "\n");
		return -devID;
	}

	cudaDeviceProp deviceProp;
	checkCudaErrors( cudaGetDeviceProperties(&deviceProp, devID) );
	if (deviceProp.major < 1) {
		fprintf(stderr, "gpuDeviceInit(): GPU device does not support CUDA.\n");
		exit(-1);                                                  \
	}

	checkCudaErrors( cudaSetDevice(devID) );
	printf("> gpuDeviceInit() CUDA device [%d]: %s\n", devID, deviceProp.name);
	return devID;
}

// This function returns the best GPU (with maximum GFLOPS)
int gpuGetMaxGflopsDeviceId()
{
	int current_device   = 0, sm_per_multiproc = 0;
	int max_compute_perf = 0, max_perf_device  = 0;
	int device_count     = 0, best_SM_arch     = 0;
	cudaDeviceProp deviceProp;

	cudaGetDeviceCount( &device_count );
	// Find the best major SM Architecture GPU device
	while ( current_device < device_count ) {
		cudaGetDeviceProperties( &deviceProp, current_device );
		if (deviceProp.major > 0 && deviceProp.major < 9999) {
			best_SM_arch = MAX(best_SM_arch, deviceProp.major);
		}
		current_device++;
	}

	// Find the best CUDA capable GPU device
	current_device = 0;
	while( current_device < device_count ) {
		cudaGetDeviceProperties( &deviceProp, current_device );
		if (deviceProp.major == 9999 && deviceProp.minor == 9999) {
			sm_per_multiproc = 1;
		} else {
			sm_per_multiproc = _ConvertSMVer2Cores(deviceProp.major, deviceProp.minor);
		}

		int compute_perf  = deviceProp.multiProcessorCount * sm_per_multiproc * deviceProp.clockRate;
		if( compute_perf  > max_compute_perf ) {
			// If we find GPU with SM major > 2, search only these
			if ( best_SM_arch > 2 ) {
				// If our device==dest_SM_arch, choose this, or else pass
				if (deviceProp.major == best_SM_arch) {
					max_compute_perf  = compute_perf;
					max_perf_device   = current_device;
				}
			} else {
				max_compute_perf  = compute_perf;
				max_perf_device   = current_device;
			}
		}
		++current_device;
	}
	return max_perf_device;
}

//    // Initialization code to find the best CUDA Device
//    int findCudaDevice(int argc, const char **argv)
//    {
//        cudaDeviceProp deviceProp;
//        int devID = 0;
//        // If the command-line has a device number specified, use it
//        if (checkCmdLineFlag(argc, argv, "device")) {
//            devID = getCmdLineArgumentInt(argc, argv, "device=");
//            if (devID < 0) {
//                printf("Invalid command line parameters\n");
//                exit(-1);
//            } else {
//                devID = gpuDeviceInit(devID);
//                if (devID < 0) {
//                   printf("exiting...\n");
//                   shrQAFinishExit(argc, (const char **)argv, QA_FAILED);
//                   exit(-1);
//                }
//            }
//        } else {
//            // Otherwise pick the device with highest Gflops/s
//            devID = gpuGetMaxGflopsDeviceId();
//            checkCudaErrors( cudaSetDevice( devID ) );
//            checkCudaErrors( cudaGetDeviceProperties(&deviceProp, devID) );
//            printf("> Using CUDA device [%d]: %s\n", devID, deviceProp.name);
//        }
//        return devID;
//    }
// end of CUDA Helper Functions


void printGPUinfo(int& block_size, int GPUindex)
{
	//Obtain GPU device info
	int deviceCount;
	cudaGetDeviceCount(&deviceCount);

	cout << "Number of GPUs = " << deviceCount << endl;
	cout << "Using GPU " << GPUindex << " (GPU index)" << endl;
	// more detailed information
//	int device;
//	for (device=0 ; device < deviceCount ; ++device) {
//		printf("\nDevice number: %d\n",device);
//		//int deviceID=GPU_USE;
//		int deviceID=device;
//		cudaDeviceProp deviceProp;
//		cudaSetDevice(deviceID);
//		char name[256];
//		cuDeviceGetName(name, 255, deviceID);
//		printf("Using device %s\n",name);
//
//		cudaGetDeviceProperties(&deviceProp, deviceID);
//		printf("totalGlobalMem %d\n", deviceProp.totalGlobalMem);
//		printf("regsPerBlock %d\n", deviceProp.regsPerBlock);
//		printf("sharedMemPerBlock %d\n", deviceProp.sharedMemPerBlock);
//		printf("maxThreadsPerBlock %d\n", deviceProp.maxThreadsPerBlock);
//		printf("maxThreadsDim %d %d %d\n", deviceProp.maxThreadsDim[0],deviceProp.maxThreadsDim[1],deviceProp.maxThreadsDim[2]);
//		printf("maxGridSize %d %d %d\n", deviceProp.maxGridSize[0],deviceProp.maxGridSize[1],deviceProp.maxGridSize[2]);
//		printf("total processor count %d\n", deviceProp.multiProcessorCount);
//		printf("warpSize %d\n", deviceProp.warpSize);
//		printf("concurrent Kernel support (this should be 1) %d\n\n", deviceProp.concurrentKernels);
//
//
//		int devID;
//		cudaDeviceProp props;
//
//		// get number of SMs on this GPU
//		checkCudaErrors(cudaGetDevice(&devID));
//		checkCudaErrors(cudaGetDeviceProperties(&props, devID));
//
//		// use a larger block size for Fermi and above
//		block_size = (props.major < 2) ? 16 : 32;
//
//		printf("Device %d: \"%s\" with Compute %d.%d capability\n", devID, props.name, props.major, props.minor);
//		cout<< "block_size : " << block_size << endl;
//	}
	//cudaSetDevice(1); // for gpufarm
	cudaSetDevice(GPUindex); // for my machine
}

void cudaPrintOn()
{
	cudaPrintfInit();
}

void cudaPrintOff()
{
	cudaPrintfEnd();
}

void cudaPrintExample()
{
	// setup execution parameters
	dim3 threads(4,1,1);
	dim3 grid(2,1,1);
	cuPrintfExample<<<grid,threads>>>();
	cudaPrintfDisplay(stdout, true);
}




// concurrent kernel execution for matrix mul. C=AB.
void concurrentMatrixMul(float* h_C, const float* h_A, const float* h_B, float* d_A, float* d_B, float* d_C)
{
	//Memory block sizes
	unsigned int a_matrix_mem_size		=	sizeof(float) * ((Nnnls * Mnnls) * conNSYS);
	unsigned int b_mem_size				=	sizeof(float) * Mnnls * conNSYS;
	unsigned int c_mem_size				=	sizeof(float) * Nnnls * conNSYS;

	// setup execution parameters
	dim3 threads(Nnnls,1,1);
	dim3 grid(conGRIDX_WIDTH,1,1);

	// copy host memory to device
	checkCudaErrors(cudaMemcpy(d_A, h_A, a_matrix_mem_size, cudaMemcpyHostToDevice));
	checkCudaErrors(cudaMemcpy(d_B, h_B, b_mem_size, cudaMemcpyHostToDevice));

	matrixMul_NNLS<<<grid, threads>>>( d_C, d_A, d_B );
	cudaThreadSynchronize();
	// check if kernel execution generated and error
	getLastCudaError("CUDA Kernel execution failed");
	cudaDeviceSynchronize();

	// copy result from device to host
	checkCudaErrors(cudaMemcpy(h_C, d_C, c_mem_size, cudaMemcpyDeviceToHost));
}


void calcSpatialNNLSprior(const float* distributions, float* distributions_prior, const int rowN,
		const int columnN, const int sliceN, float NL_control)
{
	// allocate device memory
	int mem_size_A = sizeof(float)*(rowN*columnN*sliceN*40);
	int mem_size_B = sizeof(float)*(rowN*columnN*sliceN*40);
	float* d_A, *d_B;
	checkCudaErrors(cudaMalloc((void**) &d_A, mem_size_A));
	checkCudaErrors(cudaMalloc((void**) &d_B, mem_size_B));

	// copy host memory to device
	checkCudaErrors(cudaMemcpy(d_A, distributions, mem_size_A, cudaMemcpyHostToDevice));

	// setup execution parameters
	dim3 threads(21,21,1);
	dim3 grid(rowN,columnN,sliceN);

	calcPrior_kernel<<<grid, threads>>>(d_A, d_B, rowN, columnN, sliceN, NL_control);
	// check if kernel execution generated and error
	getLastCudaError("CUDA Kernel execution failed");
	cudaPrintfDisplay(stdout, true);
	cudaDeviceSynchronize();

	// copy result from device to host
	checkCudaErrors(cudaMemcpy(distributions_prior, d_B, mem_size_B, cudaMemcpyDeviceToHost));

	checkCudaErrors(cudaFree(d_A));
	checkCudaErrors(cudaFree(d_B));
}

