
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <cmath>
#include <vector>
#include <algorithm>
#include <mkl.h>

#include "enNNLS.h"
#include "cudaKernels.h"
#include "CPU_NNLS.h"

#define PI 3.14159265

#include "headers.h"
#include "spline.hpp"
using namespace std;



//logarithmically spaced vector
void logspace(const float& d1, const float& d2, const int& n, float* T2_times)
{
	float temp;
	for (int i=0;i<=(n-2);i++) {
		temp = (float)i;
		temp *= d2 - d1;
		temp /= n - 1;
		temp += d1;
		temp = pow(10,temp);
		T2_times[i] = temp;
	}
	T2_times[n-1] = pow(10,d2);
}


//Linearly spaced vector
void linspace(const float& d1, const float& d2, const int& n, float* flip_angles)
{
	float temp;
	int n1 = n - 1;
	for (int i=0;i<n1;i++) {
		temp = (float)i;
		temp *= d2 - d1;
		temp /= n1;
		temp += d1;
		flip_angles[i] = temp;
	}
	flip_angles[0] = d1;
	flip_angles[n-1] = d2;
}


// using MKL for sparse matrix computation
void EPGdecaycurveMKL2(const int& ETL, const float& flip_angle, const float& TE, const float& T2,
		const float& T1, const float& refcon, float* echo_amp, MKL_Complex8* Tr_dia, int x, int numCPU)
{
	MKL_Complex8* MM = new MKL_Complex8[3*ETL];
	for (int i=0;i<(3*ETL);i++) {
		MM[i].real = 0;
		MM[i].imag = 0;
	}
	MM[0].real = exp(-(TE/2)/T2)*sin(flip_angle*(PI/180)/2);
	for (int i=0;i<ETL;i++) echo_amp[i] = 0;
	MKL_Complex8* T_1 = new MKL_Complex8[9]; // (3x3) matrix
	MKL_Complex8* T_p = new MKL_Complex8[(3*ETL)*(3*ETL)]; // (3*ETL)*(3*ETL) matrix
	flipmat_mkl(flip_angle*(PI/180),ETL,refcon, T_1, T_p);
	MKL_Complex8* M_tempIn = new MKL_Complex8[3];
	MKL_Complex8* M_tempOut = new MKL_Complex8[3];
	M_tempIn[0] = MM[0];
	M_tempIn[1] = MM[1];
	M_tempIn[2] = MM[2];
	MatrixMul_mkl(M_tempOut, T_1, M_tempIn, 3, 3, 1, 3);
	MM[0] = M_tempOut[0];
	MM[1] = M_tempOut[1];
	MM[2] = M_tempOut[2];
	echo_amp[0] = sqrt(pow(MM[1].real,2) + pow(MM[1].imag,2));
	echo_amp[0] = echo_amp[0]*exp(-(TE/2)/T2);

	applyRelaxationMatrixMKL2(MM, Tr_dia, T_p, echo_amp, ETL, TE, T2, x, numCPU);

	delete [] M_tempOut;
	delete [] M_tempIn;
	delete [] T_p;
	delete [] T_1;
	delete [] MM;
}


void relaxmat_mkl_preComputed(const int& ETL, const float& TE, const float& T2,
		const float& T1, MKL_Complex8* T_r, int angleInd)
{
	int Tr_memsize = (3*ETL)*(3*ETL);
	MKL_Complex8 Tr_temp[Tr_memsize];
	copy(T_r + Tr_memsize*angleInd, T_r + Tr_memsize*(angleInd+1), Tr_temp);
	// initialize T_r
	for (int x=0;x<Tr_memsize;x++) {
		Tr_temp[x].real = 0;
		Tr_temp[x].imag = 0;
	}
	Tr_temp[IDX2C(1,0,3*ETL)].real = exp(-TE/T2);
	Tr_temp[IDX2C(1,0,3*ETL)].imag = 0;
	int i,j;
	for (int x=0;x<(ETL-1);x++) {
		j = 3*x + 1;
		i = 3*x + 4;
		Tr_temp[IDX2C(i,j,3*ETL)].real = exp(-TE/T2);
		Tr_temp[IDX2C(i,j,3*ETL)].imag = 0;
	}
	for (int x=0;x<(ETL-1);x++) {
		j = 3*x + 3;
		i = 3*x;
		Tr_temp[IDX2C(i,j,3*ETL)].real = exp(-TE/T2);
		Tr_temp[IDX2C(i,j,3*ETL)].imag = 0;
	}
	for (int x=0;x<ETL;x++) {
		j = 3*x + 2;
		i = 3*x + 2;
		Tr_temp[IDX2C(i,j,3*ETL)].real = exp(-TE/T1);
		Tr_temp[IDX2C(i,j,3*ETL)].imag = 0;
	}
	copy(Tr_temp, Tr_temp + Tr_memsize, T_r + Tr_memsize*angleInd);
}


void flipmat_mkl(const float& alpha, const int& num_pulses, const float& refcon, MKL_Complex8* T_1, MKL_Complex8* T_p)
{
	T_1[0].real = pow(cos(alpha/2),2);
	T_1[0].imag = 0;
	T_1[1].real = pow(sin(alpha/2),2);
	T_1[1].imag = 0;
	T_1[2].real = 0;
	T_1[2].imag = -sin(alpha);
	T_1[3].real = pow(sin(alpha/2),2);
	T_1[3].imag = 0;
	T_1[4].real = pow(cos(alpha/2),2);
	T_1[4].imag = 0;
	T_1[5].real = 0;
	T_1[5].imag = sin(alpha);
	T_1[6].real = 0;
	T_1[6].imag = -0.5*sin(alpha);
	T_1[7].real = 0;
	T_1[7].imag = 0.5*sin(alpha);
	T_1[8].real = cos(alpha);
	T_1[8].imag = 0;

	float alpha2 = alpha*refcon/180;
	MKL_Complex8 T_2[9];

	T_2[0].real = pow(cos(alpha2/2),2);
	T_2[0].imag = 0;
	T_2[1].real = pow(sin(alpha2/2),2);
	T_2[1].imag = 0;
	T_2[2].real = 0;
	T_2[2].imag = -sin(alpha2);
	T_2[3].real = pow(sin(alpha2/2),2);
	T_2[3].imag = 0;
	T_2[4].real = pow(cos(alpha2/2),2);
	T_2[4].imag = 0;
	T_2[5].real = 0;
	T_2[5].imag = sin(alpha2);
	T_2[6].real = 0;
	T_2[6].imag = -0.5*sin(alpha2);
	T_2[7].real = 0;
	T_2[7].imag = 0.5*sin(alpha2);
	T_2[8].real = cos(alpha2);
	T_2[8].imag = 0;

	// initialize T_p
	for (int x=0;x<((3*num_pulses)*(3*num_pulses));x++) {
		T_p[x].real = 0;
		T_p[x].imag = 0;
	}

	for (int x=0;x<num_pulses;x++) {
		T_p[IDX2C(3*x,3*x,3*num_pulses)] = T_2[0];
		T_p[IDX2C(3*x+1,3*x,3*num_pulses)] = T_2[1];
		T_p[IDX2C(3*x+2,3*x,3*num_pulses)] = T_2[2];
		T_p[IDX2C(3*x,3*x+1,3*num_pulses)] = T_2[3];
		T_p[IDX2C(3*x+1,3*x+1,3*num_pulses)] = T_2[4];
		T_p[IDX2C(3*x+2,3*x+1,3*num_pulses)] = T_2[5];
		T_p[IDX2C(3*x,3*x+2,3*num_pulses)] = T_2[6];
		T_p[IDX2C(3*x+1,3*x+2,3*num_pulses)] = T_2[7];
		T_p[IDX2C(3*x+2,3*x+2,3*num_pulses)] = T_2[8];
	}
}


// solve Ax=b using regularization
void NNLSreg(REAL* h_A, REAL* h_b, REAL* h_x, const float& Chi2Factor, int numCPU)
{
	//Set number of threads
	mkl_set_num_threads(numCPU);

	REAL X_noreg[Mnnls];
	REAL b_backprojected[Nnnls];
	/*
	 * Find non-regularized solution
	 */
	nnlsOMPSysMKLUpdates(h_A, h_b, X_noreg, 1, MAX_ITER_NNLS_CPU(32,40),
			MAX_ITER_LS_CPU(32,40), 1, 32, 40, numCPU, 1, 1e-6); // CPU NNLS

#ifndef USE_DOUBLE
	cblas_sgemv(CblasRowMajor, CblasNoTrans, Nnnls, Mnnls, 1.0, h_A, Mnnls, X_noreg, 1, 0.0, b_backprojected, 1);
#else
	cblas_dgemv(CblasRowMajor, CblasNoTrans, Nnnls, Mnnls, 1.0, h_A, Mnnls, X_noreg, 1, 0.0, b_backprojected, 1);
#endif

	REAL chi2_min=0;
	for (int i=0 ; i<Nnnls ; i++) {
		chi2_min += (h_b[i] - b_backprojected[i])*(h_b[i] - b_backprojected[i]);
	}

	vector<REAL> mu;
	mu.push_back(0.0);
	vector<REAL> chi2;
	chi2.push_back(chi2_min);
	REAL H[Mnnls*Mnnls] = {0}; // Mnnls x Mnnls matrix
	for (int i=0 ; i<Mnnls ; i++) {
		H[IDX2C(i,i,Mnnls)] = 1;
	}
	/*
	 * Loop to find largest mu that keeps chi-squared in desired range
	 */
	REAL smooth[Mnnls*Mnnls] = {0}; // Mnnls x Mnnls matrix
	REAL Csmooth[Mnnls*Nreg]; // Nreg x Mnnls matrix
	REAL X_reg[Mnnls];
	REAL breg[Nreg];
	for (int i=0 ; i<Nreg ; i++) {
		if (i<Nnnls) {
			breg[i] = h_b[i];
		} else {
			breg[i] = 0;
		}
	}
	REAL chi2_temp;

	while (chi2.back() < ((REAL)Chi2Factor*chi2_min)) {
		// Increment mu vector
		if (mu.back() > 0.0) {
			mu.push_back(2*(mu.back()));
		} else {
			mu.push_back(0.001);
		}
		// Compute T2 distribution with smoothing
		for (int i=0 ; i<Mnnls ; i++) {
			smooth[IDX2C(i,i,Mnnls)] = (mu.back())*H[IDX2C(i,i,Mnnls)];
		}
		memcpy(Csmooth, h_A, (Nnnls*Mnnls)*sizeof(REAL));
		memcpy(&Csmooth[Nnnls*Mnnls], smooth, (Mnnls*Mnnls)*sizeof(REAL));

		nnlsOMPSysMKLUpdates(Csmooth, breg, X_reg, 1, MAX_ITER_NNLS_CPU(72,40),
				MAX_ITER_LS_CPU(72,40), 1, 72, 40, numCPU, 1, 4e-10); // CPU NNLS

		// Find predicted curve and calculate residuals and chi-squared
#ifndef USE_DOUBLE
		cblas_sgemv(CblasRowMajor, CblasNoTrans, Nnnls, Mnnls, 1.0, h_A, Mnnls, X_reg, 1, 0.0, b_backprojected, 1);
#else
		cblas_dgemv(CblasRowMajor, CblasNoTrans, Nnnls, Mnnls, 1.0, h_A, Mnnls, X_reg, 1, 0.0, b_backprojected, 1);
#endif
		chi2_temp = 0;
		for (int i=0 ; i<Nnnls ; i++) {
			chi2_temp += (h_b[i] - b_backprojected[i])*(h_b[i] - b_backprojected[i]);
		}
		chi2.push_back(chi2_temp);
	}

	/*
	 * Smooth the chi2(mu) curve using spline fit
	 */
	using namespace magnet::math;
	Spline sp;
	for (int i=0; i<(int)mu.size(); i++) {
		sp.addPoint(mu[i], chi2[i]);
	}
	vector<REAL> mu_spline;
	vector<REAL> chi2_spline;
	int pos;
	// find the minimum chi-squared and the corresponding angle
	for (REAL a=0; a<=(mu.back()+0.001) ; a+=0.001) {
		mu_spline.push_back(a);
		chi2_spline.push_back(sp(a));
	}
	for (int i=0 ; i<(int)chi2_spline.size() ; i++) {
		chi2_spline[i] = fabs(chi2_spline[i] - ((REAL)Chi2Factor*chi2_min));
	}
	pos = min_element(chi2_spline.begin(), chi2_spline.end()) - chi2_spline.begin();
	REAL mu2 = mu_spline[pos];
	for (int i=0 ; i<Mnnls ; i++) {
		smooth[IDX2C(i,i,Mnnls)] = mu2*H[IDX2C(i,i,Mnnls)];
	}
	// Compute the regularized solution
	memcpy(Csmooth, h_A, (Nnnls*Mnnls)*sizeof(REAL));
	memcpy(&Csmooth[Nnnls*Mnnls], smooth, (Mnnls*Mnnls)*sizeof(REAL));
	nnlsOMPSysMKLUpdates(Csmooth, breg, X_reg, 1, MAX_ITER_NNLS_CPU(72,40),
			MAX_ITER_LS_CPU(72,40), 1, 72, 40, numCPU, 1, 4e-10); // CPU NNLS
	memcpy(h_x, X_reg, sizeof(REAL)*Mnnls);
}


void srNNLS(REAL* h_A, REAL* h_b, REAL* h_x, const float& Chi2Factor,
		REAL* priorT2, float reg_control, int numCPU)
{
	//Set number of threads
	mkl_set_num_threads(numCPU);

	REAL X_noreg[Mnnls];
	REAL b_backprojected[Nnnls];
	/*
	 * Find non-regularized solution
	 */
	nnlsOMPSysMKLUpdates(h_A, h_b, X_noreg, 1, MAX_ITER_NNLS_CPU(32,40),
			MAX_ITER_LS_CPU(32,40), 1, 32, 40, numCPU, 1, 1e-6); // CPU NNLS
#ifndef USE_DOUBLE
	cblas_sgemv(CblasRowMajor, CblasNoTrans, Nnnls, Mnnls, 1.0, h_A, Mnnls, X_noreg, 1, 0.0, b_backprojected, 1);
#else
	cblas_dgemv(CblasRowMajor, CblasNoTrans, Nnnls, Mnnls, 1.0, h_A, Mnnls, X_noreg, 1, 0.0, b_backprojected, 1);
#endif
	REAL chi2_min=0;
	for (int i=0 ; i<Nnnls ; i++) {
		chi2_min += (h_b[i] - b_backprojected[i])*(h_b[i] - b_backprojected[i]);
	}

	vector<REAL> mu;
	mu.push_back(0.0);
	vector<REAL> chi2;
	chi2.push_back(chi2_min);
	REAL H[Mnnls*Mnnls] = {0}; // Mnnls x Mnnls matrix
	for (int i=0 ; i<Mnnls ; i++) {
		H[IDX2C(i,i,Mnnls)] = 1;
	}
	/*
	 * Loop to find largest mu that keeps chi-squared in desired range
	 */
	REAL smooth[Mnnls*Mnnls] = {0}; // Mnnls x Mnnls matrix
	REAL Csmooth[Mnnls*Nreg]; // Nreg x Mnnls matrix
	REAL X_reg[Mnnls];
	REAL breg[Nreg];
	for (int i=0 ; i<Nreg ; i++) {
		if (i<Nnnls) {
			breg[i] = h_b[i];
		} else {
			breg[i] = priorT2[i-Nnnls];
		}
	}
	float chi2_temp;

	REAL chi2_pre = 0.0;
	int itrLimit = 15;
	int itrC = 0;
	while ((chi2.back() < ((REAL)reg_control*chi2_min)) && (chi2.back() > chi2_pre) && (itrC < itrLimit)) {
		chi2_pre = chi2.back();
		// Increment mu vector
		if (mu.back() > 0.0) {
			mu.push_back(2*(mu.back()));
		} else {
			mu.push_back(0.01);
		}
		// Compute T2 distribution with constraints
		for (int i=0 ; i<Mnnls ; i++) {
			smooth[IDX2C(i,i,Mnnls)] = (mu.back())*H[IDX2C(i,i,Mnnls)];
		}
		memcpy(Csmooth, h_A, (Nnnls*Mnnls)*sizeof(REAL));
		memcpy(&Csmooth[Nnnls*Mnnls], smooth, (Mnnls*Mnnls)*sizeof(REAL));
		for (int i=Nnnls ; i<Nreg ; i++) breg[i] = mu.back()*priorT2[i-Nnnls];
		nnlsOMPSysMKLUpdates(Csmooth, breg, X_reg, 1, MAX_ITER_NNLS_CPU(72,40),
				MAX_ITER_LS_CPU(72,40), 1, 72, 40, numCPU, 1, 4e-10); // CPU NNLS

		// Find predicted curve and calculate residuals and chi-squared
#ifndef USE_DOUBLE
		cblas_sgemv(CblasRowMajor, CblasNoTrans, Nnnls, Mnnls, 1.0, h_A, Mnnls, X_reg, 1, 0.0, b_backprojected, 1);
#else
		cblas_dgemv(CblasRowMajor, CblasNoTrans, Nnnls, Mnnls, 1.0, h_A, Mnnls, X_reg, 1, 0.0, b_backprojected, 1);
#endif
		chi2_temp = 0;
		for (int i=0 ; i<Nnnls ; i++) {
			chi2_temp += (h_b[i] - b_backprojected[i])*(h_b[i] - b_backprojected[i]);
		}
		chi2.push_back(chi2_temp);
		itrC++;
	}
	/*
	 * Smooth the chi2(mu) curve using spline fit
	 */
	using namespace magnet::math;
	Spline sp;
	for (int i=0 ; i<(int)mu.size() ; i++) {
		sp.addPoint(mu[i], chi2[i]);
	}
	vector<REAL> mu_spline;
	vector<REAL> chi2_spline;
	int pos;
	// find the minimum chi-squared and the corresponding angle
	for (REAL a=0; a<=(mu.back()+0.001) ; a+=0.001) {
		mu_spline.push_back(a);
		chi2_spline.push_back(sp(a));
	}
	for (int i=0 ; i<(int)chi2_spline.size() ; i++) {
		chi2_spline[i] = fabs(chi2_spline[i] - ((REAL)reg_control*chi2_min));
	}
	pos = min_element(chi2_spline.begin(), chi2_spline.end()) - chi2_spline.begin();

	REAL mu2 = mu_spline[pos];
	for (int i=0 ; i<Mnnls ; i++) {
		smooth[IDX2C(i,i,Mnnls)] = mu2*H[IDX2C(i,i,Mnnls)];
	}
	// Compute the regularized solution
	memcpy(Csmooth, h_A, (Nnnls*Mnnls)*sizeof(REAL));
	memcpy(&Csmooth[Nnnls*Mnnls], smooth, (Mnnls*Mnnls)*sizeof(REAL));
	for (int i=Nnnls ; i<Nreg ; i++) breg[i] = mu2*priorT2[i-Nnnls];
	nnlsOMPSysMKLUpdates(Csmooth, breg, X_reg, 1, MAX_ITER_NNLS_CPU(72,40),
			MAX_ITER_LS_CPU(72,40), 1, 72, 40, numCPU, 1, 4e-10); // CPU NNLS
	memcpy(h_x, X_reg, sizeof(REAL)*Mnnls);
}


void MatrixMul_mkl(MKL_Complex8* C, MKL_Complex8* A, MKL_Complex8* B,
		unsigned int uiWA, unsigned int uiHA, unsigned int uiWB, unsigned int uiHB)
{
	unsigned int uiWC = uiWB; // # of columns
	unsigned int uiHC = uiHA; // # of rows
	MKL_Complex8 temp;
	for (unsigned int j=0;j<uiHC;j++) {
		for (unsigned int i=0;i<uiWC;i++) {
			temp.real = 0;
			temp.imag = 0;
			for (unsigned int k=0;k<uiWA;k++) {
				temp.real += A[IDX2C(k,j,uiWA)].real*B[IDX2C(i,k,uiWB)].real - A[IDX2C(k,j,uiWA)].imag*B[IDX2C(i,k,uiWB)].imag;
				temp.imag += A[IDX2C(k,j,uiWA)].real*B[IDX2C(i,k,uiWB)].imag + A[IDX2C(k,j,uiWA)].imag*B[IDX2C(i,k,uiWB)].real;
			}
			C[IDX2C(i,j,uiWC)] = temp;
		}
	}
}


void applyRelaxationMatrixMKL2(MKL_Complex8* MM, MKL_Complex8* Tr_dia_org, MKL_Complex8* T_p, float* echo_amp,
		const int& ETL, const float& TE, const float& T2, int x, int numCPU)
{
	//Set number of threads
	mkl_set_num_threads(numCPU);

	MKL_INT M=96;
	MKL_INT N=96;
	int nnzTp=288;
	MKL_INT info;
	MKL_INT job[6];
	/*
	 * convert T_P to the CSR format
	 */
	job[0] = 0;
	job[1] = 1;
	job[2] = 1;
	job[3] = 2;
	job[4] = nnzTp;
	job[5] = 1;
	MKL_Complex8 Tp_csr[nnzTp]; // NZMAX
	MKL_INT Tp_I[97]; // number of row + 1
	MKL_INT Tp_J[nnzTp]; // NZMAX
	mkl_cdnscsr(job, &M, &N, T_p, &M, Tp_csr, Tp_J, Tp_I, &info);
	/*
	 * convert into the diagonal format
	 */
	job[0] = 0;
	job[5] = 0;
	MKL_INT idiagTr = 4;
	MKL_INT ndiagTr = 96;
	MKL_INT indiaTr = idiagTr*ndiagTr;
	MKL_Complex8 Tr_dia[indiaTr];
	copy(Tr_dia_org + indiaTr*x, Tr_dia_org + indiaTr*(x+1), Tr_dia);
	MKL_INT Tr_distance[idiagTr];
	Tr_distance[0] = -3;
	Tr_distance[1] = -1;
	Tr_distance[2] = 0;
	Tr_distance[3] = 3;
	int dim = 96;
	MKL_INT idiagTp = 5;
	MKL_INT ndiagTp = 96;
	MKL_INT indiaTp = idiagTp*ndiagTp;
	MKL_Complex8 Tp_dia[indiaTp];
	MKL_INT Tp_distance[idiagTp];
	Tp_distance[0] = -2;
	Tp_distance[1] = -1;
	Tp_distance[2] = 0;
	Tp_distance[3] = 1;
	Tp_distance[4] = 2;
	mkl_ccsrdia(job, &dim, Tp_csr, Tp_J, Tp_I, Tp_dia, &ndiagTp, Tp_distance, &idiagTp,
			Tp_csr, Tp_J, Tp_I, &info);
	MKL_Complex8* MM_temp = new MKL_Complex8[M];
	char transa = 'T';
	mkl_cdiagemv(&transa, &M, Tr_dia, &M, Tr_distance, &idiagTr, MM, MM_temp);
//#pragma unroll
	// perform flip-relax sequence ETL-1 times
	for (int x=1 ; x<ETL ; x++) {
		// perform the flip
		mkl_cdiagemv(&transa, &M, Tp_dia, &M, Tp_distance, &idiagTp, MM_temp, MM);
		// Record the magnitude of the population of F1* as the echo amplitude and allow for relaxation
		echo_amp[x] = sqrt(pow(MM[1].real,2) + pow(MM[1].imag,2))*exp(-(TE/2)/T2);
		// Allow time evolution of magnetization between pulses
		mkl_cdiagemv(&transa, &M, Tr_dia, &M, Tr_distance, &idiagTr, MM, MM_temp);
	}
	delete [] MM_temp;
}


// pre-compute Tr_dia
void calculate_Tr_dia(MKL_Complex8* T_r, MKL_Complex8* Tr_dia, int x)
{
	int ETL = 32;
	int Tr_memsize = (3*ETL)*(3*ETL);
	MKL_Complex8 Tr_temp[Tr_memsize];
	copy(T_r + Tr_memsize*x, T_r + Tr_memsize*(x+1), Tr_temp);

	MKL_INT M=96;
	MKL_INT N=96;
	int nnzTr=95;
	MKL_INT info;
	MKL_INT job[6];
	/*
	 * convert T_r into the CSR format
	 */
	job[0] = 0;
	job[1] = 1;
	job[2] = 1;
	job[3] = 2;
	job[4] = nnzTr;
	job[5] = 1;
	MKL_Complex8 Tr_csr[nnzTr];
	MKL_INT Tr_I[97]; // number of row + 1
	MKL_INT Tr_J[nnzTr]; // NZMAX
	mkl_cdnscsr(job, &M, &N, Tr_temp, &M, Tr_csr, Tr_J, Tr_I, &info);
	/*
	 * convert into the diagonal format
	 */
	job[0] = 0;
	job[5] = 0;
	MKL_INT idiagTr = 4;
	MKL_INT ndiagTr = 96;
	MKL_INT indiaTr = idiagTr*ndiagTr;
	MKL_Complex8 Tr_dia_temp[indiaTr];
	MKL_INT Tr_distance[idiagTr];
	Tr_distance[0] = -3;
	Tr_distance[1] = -1;
	Tr_distance[2] = 0;
	Tr_distance[3] = 3;
	int dim = 96;
	mkl_ccsrdia(job, &dim, Tr_csr, Tr_J, Tr_I, Tr_dia_temp, &ndiagTr, Tr_distance, &idiagTr,
			Tr_csr, Tr_J, Tr_I, &info);
	copy(Tr_dia_temp, Tr_dia_temp + indiaTr, Tr_dia + indiaTr*x);
}


