/*
 * T2 relaxation data analysis with stimulated echo correction and non-local spatial regularization
 *
 * This work has been published to
 *
 * Journal of Magnetic Resonance Imaging (2014):
 * Fast Computation of Myelin Maps from MRI T2 Relaxation Data Using Multicore CPU and Graphics Card Parallelization
 *
 * Medical Image Computing and Computer Assisted Intervention (2013):
 * Non-local spatial regularization of MRI T2 relaxation images for myelin water quantification
 *
 *
 * Written by Youngjin Yoo, University of British Columbia, 2013.
 */
#include <cstdio>
#include <cstdlib>
#include <vector>
#include <cstring>
#include <fstream>
#include <iostream>
#include <memory>
#include <ctime>
#include <algorithm>
#include "omp.h"
#include <mkl.h>
#include <typeinfo>
#include <cuda_runtime.h>
#include <cmath>
#include "enNNLS.h"
#include "CPU_NNLS.h"
#include "cudaKernels.h"
#include "spline.hpp"
#include "headers2.h"
#include "nifti1.h"

using namespace std;

int main(int argc, char ** argv)
{	
	// Verify the number of parameters in the command line
	int NLSR_flag = 0; // non-local spatial regularization, off=0, on=1
	if (argc == 6) {
		NLSR_flag = 1;
	} else if ((argc != 4) && (argc != 6)) {
		cerr << "Missing Parameters " << endl;
		cerr << endl;
		cerr << "Without non-local spatial regularization: " << endl;
		cerr << "Usage: " << argv[0];
		cerr << " [root T2multiechoFile (without echo number)] ";
		cerr << " [output MWF map filename] ";
		cerr << " [GPU index] " << endl;
		cerr << endl;
		cerr << "With non-local spatial regularization: " << endl;
		cerr << "Usage: " << argv[0];
		cerr << " [root T2multiechoFile (without echo number)] ";
		cerr << " [output MWF map filename] ";
		cerr << " [NL-means (default: 1.5)] ";
		cerr << " [L2 regularization (default: 1.2)] ";
		cerr << " [GPU index] " << endl;
		return EXIT_FAILURE;
	}

	cout << "" << endl;
	cout << "***************************************************" << endl;
	cout << "* Myelin water map estimation using nlsrNNLS" << endl;
	cout << "***************************************************" << endl;
	cout << "" << endl;

	// find the number of available CPU cores for Intel MKL
	int numCPU = sysconf( _SC_NPROCESSORS_ONLN );
	cout << "number of available CPUs = " << numCPU << endl;
	cout << "will use " << numCPU << " threads for Intel MKL..." << endl;

	// find the number of available CPU cores for Intel MKL
	cout << endl;
	cout << "Using REAL = " << typeid(REAL).name() << " for PBLAS..." << endl;
	cout << endl;

	// show GPU info and get block_size
	int block_size;
	int GPIindex = atoi(argv[5]);
	printGPUinfo(block_size, GPIindex);
	cudaPrintOn();

	// load input images
	string rootInputFileName(argv[1]);
	if (rootInputFileName == "") {
		cerr << "Error: must specify input images." << endl;
		return EXIT_FAILURE;
	} else {
		cout << "\nroot T2 echo input file name : " << rootInputFileName << endl;
	}
	
	int echoN = 32; // number of echoes

	// nii images variables
	nifti_1_header hdr;
	FILE *fp;
	int ret,i;
	nifti1_extender pad={0,0,0,0};

	// get image information
	string filename1;
	string filename2;
	stringstream s1;
	string ss1;
	s1.str("");
	s1 << 1;
	ss1 = s1.str();

	// for the case echo number in file name has 0x pattern
	// otherwise just use: filename1 = ss1;
	filename1 = "0";
	filename1 += ss1;
	filename1 += ".nii";
	filename2 = rootInputFileName;
	filename2 += filename1;

	/********** open and read header */
	fp = fopen(filename2.c_str(),"r");
	if (fp == NULL) {
		fprintf(stderr, "\nError opening header file %s\n",filename2.c_str());
		exit(1);
	}
	ret = fread(&hdr, MIN_HEADER_SIZE, 1, fp);
	if (ret != 1) {
		fprintf(stderr, "\nError reading header file %s\n",filename2.c_str());
		exit(1);
	}
	fclose(fp);

	int rowN, columnN, sliceN;
	rowN = hdr.dim[1];
	columnN = hdr.dim[2];
	sliceN = hdr.dim[3];

	float* T2_decay_map = new float[rowN*columnN*sliceN*echoN];

	// load entire T2 decay data
	NII_DATATYPE *nii_data = NULL;
	nii_data = (NII_DATATYPE *) malloc(sizeof(NII_DATATYPE) * hdr.dim[1]*hdr.dim[2]*hdr.dim[3]);
	if (nii_data == NULL) {
		fprintf(stderr, "\nError allocating data buffer for %s\n",filename2.c_str());
		exit(1);
	}
	float *nii_data_float = NULL;
	nii_data_float = (float *) malloc(sizeof(float) * hdr.dim[1]*hdr.dim[2]*hdr.dim[3]);
	if (nii_data_float == NULL) {
		fprintf(stderr, "\nError allocating data buffer for %s\n",filename2.c_str());
		exit(1);
	}

	int echoCount;
	for (echoCount=1;echoCount<=echoN;echoCount++) {
		s1.str("");
		s1 << echoCount;
		ss1 = s1.str();

		// for the case echo number in file name has 0x pattern
		// otherwise just use: filename1 = ss1;
		if (echoCount < 10) {
			filename1 = "0";
			filename1 += ss1;
		} else {
			filename1 = ss1;
		}

		filename1 += ".nii";
		filename2 = rootInputFileName;
		filename2 += filename1;

		// load nii image
		/********** open and read header */
		fp = fopen(filename2.c_str(),"r");
		if (fp == NULL) {
			fprintf(stderr, "\nError opening header file %s\n",filename2.c_str());
			exit(1);
		}
		ret = fread(&hdr, MIN_HEADER_SIZE, 1, fp);
		if (ret != 1) {
			fprintf(stderr, "\nError reading header file %s\n",filename2.c_str());
			exit(1);
		}
		fclose(fp);

		/********** open the datafile, jump to data offset */
		fp = fopen(filename2.c_str(),"r");
		if (fp == NULL) {
			fprintf(stderr, "\nError opening data file %s\n",filename2.c_str());
			exit(1);
		}

		ret = fseek(fp, (long)(hdr.vox_offset), SEEK_SET);
		if (ret != 0) {
			fprintf(stderr, "\nError doing fseek() to %ld in data file %s\n",(long)(hdr.vox_offset), filename2.c_str());
			exit(1);
		}

		/********** read first 3D volume from data file */
		ret = fread(nii_data, sizeof(NII_DATATYPE), hdr.dim[1]*hdr.dim[2]*hdr.dim[3], fp);
		if (ret != hdr.dim[1]*hdr.dim[2]*hdr.dim[3]) {
			fprintf(stderr, "\nError reading volume 1 from %s (%d)\n",filename2.c_str(),ret);
			exit(1);
		}
		fclose(fp);

		/********** scale the data buffer  */
		if (hdr.scl_slope != 0) {
			for (i=0; i<hdr.dim[1]*hdr.dim[2]*hdr.dim[3]; i++)
				nii_data_float[i] = ((float)nii_data[i]*hdr.scl_slope) + hdr.scl_inter;
		}

		// copy decay data to buffer
		memcpy(&T2_decay_map[rowN*columnN*sliceN*(echoCount-1)], nii_data_float, (rowN*columnN*sliceN)*sizeof(float));
	}

	// for timer
	double elapsed_secs;
	clock_t begin;
	clock_t end;

	// find the basis matrices for each flip angle
	float d1 = 0.015;
	float d2 = 2;
	int nT2 = 40;
	int nechs = echoN;
	int nangles = 8;
	float minangle = 50;
	float TE = 0.01;
	float T1 = 1;
	float RefCon = 180;
	float Threshold = 200;
	float Chi2Factor = 1.02;
	float T2_times[nT2];
	float flip_angles[nangles];

	// Initialize parameters and variable for angle optimization
	logspace(log10(d1), log10(d2), nT2, T2_times);
	linspace(minangle, 180, nangles, flip_angles);
	float* basis_angles_single = new float[nT2*nechs*nangles]; // (32x40 matrix)
	float* echo_amp = new float[nechs];
	REAL* basis_angles = new REAL[nT2*nechs*nangles]; // (32x40 matrix)
	REAL* basis_decay = new REAL[nT2*nechs];

	/*
	 * Loop to compute each basis and assign them to a cell in the array
	 */
	// pre-compute T_r and Tr_dia matrix according to T2_times
	MKL_INT idiagTr = 4;
	MKL_INT ndiagTr = 96;
	MKL_INT indiaTr = idiagTr*ndiagTr;
	MKL_Complex8* T_r = new MKL_Complex8[(3*nechs)*(3*nechs)*nT2];
	MKL_Complex8* Tr_dia = new MKL_Complex8[indiaTr*nT2];
	for (int x=0 ; x<nT2 ; x++) {
		relaxmat_mkl_preComputed(nechs, TE, T2_times[x], T1, T_r, x);
		calculate_Tr_dia(T_r, Tr_dia, x);
	}
	cout << endl;
	cout << "computing basis matrix for each flip angle..." << endl;
	begin = clock();
	for (int a=0 ; a<nangles ; a++) {
		for (int x=0 ; x<nT2 ; x++) {
			EPGdecaycurveMKL2(nechs, flip_angles[a], TE, T2_times[x], T1, RefCon, echo_amp, Tr_dia, x, numCPU);
			for (int y=0; y<nechs ; y++) basis_decay[IDX2C(x,y,nT2)] = (REAL)echo_amp[y];
		}
		for (int x=0 ; x<nT2 ; x++) {
			for (int y=0; y<nechs ; y++) {
				basis_angles[IDX3C(x,y,a,nT2,nechs)] = basis_decay[IDX2C(x,y,nT2)];
				basis_angles_single[IDX3C(x,y,a,nT2,nechs)] = (float)basis_decay[IDX2C(x,y,nT2)];
			}
		}
	}
	end = clock();
	elapsed_secs = double(end - begin) / CLOCKS_PER_SEC;
	cout << "Time elapsed: " << elapsed_secs << " seconds"<< endl;
	delete [] T_r;

	/*
	 * Process all pixels
	 * Main triple for-loop to run through each pixel in the slice
	 */
	float* distributions = new float[columnN*rowN*sliceN*nT2];
	float* distributions_prior = new float[columnN*rowN*sliceN*nT2]; // for spatial prior
	float* alpha = new float[columnN*rowN*sliceN];
	float chi2_alpha[nangles];
	REAL* decay_data = new REAL[nechs];
	REAL* concurrentDecay_data = new REAL[nangles*nechs]; // nangles x nechs matrix
	REAL* T2_dis_ls = new REAL[nT2];
	float* concurrentT2_dis_ls_single = new float[nangles*nT2]; // nangles x nT2 matrix, single precision
	REAL* concurrentT2_dis_ls = new REAL[nangles*nT2]; // nangles x nT2 matrix
	float* decay_pred = new float[nangles*nechs]; // nangles x nechs matrix
	REAL* priorT2 = new REAL[nT2]; // for non-local regularization

	// cuda device memory malloc
	int Ndim = 32;
	int Mdim = 40;
	int NSYSdim = 8;

	// Memory block sizes
	unsigned int am_matrix_mem_size		=	sizeof(float) * ((Ndim * Mdim) * NSYSdim);
	unsigned int bm_mem_size				=	sizeof(float) * Mdim * NSYSdim;
	unsigned int cm_mem_size				=	sizeof(float) * Ndim * NSYSdim;

	// allocate device memory
	float* d_Am, *d_Bm, *d_Cm;
	cudaMalloc((void**) &d_Am, am_matrix_mem_size);
	cudaMalloc((void**) &d_Bm, bm_mem_size);
	cudaMalloc((void**) &d_Cm, cm_mem_size);

	// for Boost Spline interpolation
	float flip_angles_inc = 1;
	using namespace magnet::math;
	Spline spline;
	vector<float> flip_spline;
	vector<float> chi2_spline;
	int pos;

	/*
	 * starts T2 relaxation data analysis
	 */
	int debug_flag;
	cout << endl;
	cout << "calculating initial T2 distributions..." << endl;
	begin = clock();
	for (int row=0; row<rowN; row++) {
		// main loop
		for (int col=0; col<columnN; col++) {
			for (int slice=0; slice<sliceN; slice++) {
				// initialize dist.
				for (int x=0 ; x<nT2 ; x++) {
					distributions[IDX4C(row,col,slice,x,rowN,columnN,sliceN)] = 0;
					distributions_prior[IDX4C(row,col,slice,x,rowN,columnN,sliceN)] = 0;
				}
				// conditional loop to reject low signal pixels
				if (T2_decay_map[slice*(rowN*columnN) + columnN*row + col] >= Threshold) {
					for (int i=0 ; i<nechs ; i++) {
						decay_data[i] = (REAL)T2_decay_map[i*(sliceN*rowN*columnN) + slice*(rowN*columnN) + columnN*row + col];
						// for concurrent processing
						for (int j=0; j<nangles ; j++) {
							concurrentDecay_data[IDX2C(i,j,nechs)] = decay_data[i];
						}
					}

					nnlsOMPSysMKLUpdates(basis_angles, concurrentDecay_data, concurrentT2_dis_ls, 1,
							MAX_ITER_NNLS_CPU(32,40), MAX_ITER_LS_CPU(32,40), 8, 32, 40, numCPU, 1, 1e-6);
#ifndef USE_DOUBLE
					concurrentMatrixMul(decay_pred, basis_angles_single, concurrentT2_dis_ls,
							d_Am, d_Bm, d_Cm);
#else
					for (int i=0; i<(nangles*nT2); ++i)
						concurrentT2_dis_ls_single[i] = (float)concurrentT2_dis_ls[i];
					concurrentMatrixMul(decay_pred, basis_angles_single, concurrentT2_dis_ls_single, d_Am, d_Bm, d_Cm);
#endif
					/*
					 * Spline interpolation
					 */
					for (int i = 0; i < nangles; i++) {
						chi2_alpha[i] = 0;
						for (int j = 0; j < nechs; j++) {
							chi2_alpha[i] += ((float)decay_data[j] - decay_pred[IDX2C(j,i,nechs)])
									* ((float)decay_data[j] - decay_pred[IDX2C(j,i,nechs)]);
						}
					}
					spline.clear();
					for (int i = 0; i < nangles; i++) {
						spline.addPoint(flip_angles[i], chi2_alpha[i]);
					}
					flip_spline.clear();
					chi2_spline.clear();
					for (float a = flip_angles[0]; a <= flip_angles[nangles - 1]; a += flip_angles_inc) {
						flip_spline.push_back(a);
						chi2_spline.push_back(spline(a));
					}
					// find the minimum chi-squared and the corresponding angle
					pos = min_element(chi2_spline.begin(), chi2_spline.end()) - chi2_spline.begin();
					alpha[IDX3C(row,col,slice,rowN,columnN)] = flip_spline[pos];

					/*
					 * Compute the NNLS basis over T2 space
					 */
					for (int x=0 ; x<nT2 ; x++) {
						EPGdecaycurveMKL2(nechs, alpha[IDX3C(row,col,slice,rowN,columnN)], TE, T2_times[x],
								T1, RefCon, echo_amp, Tr_dia, x, numCPU);
						for (int y=0; y<nechs ; y++) {
							basis_decay[IDX2C(x,y,nT2)] = (REAL)echo_amp[y];
						}
					}

					/*
					 * Fit T2 distribution using NNLS (with L2 regularization)
					 */
					NNLSreg(basis_decay, decay_data, T2_dis_ls, Chi2Factor, numCPU);

					for (int x=0 ; x<nT2 ; x++) {
						distributions[IDX4C(row,col,slice,x,rowN,columnN,sliceN)] = (float)T2_dis_ls[x];
					}
				}
			}
		}
	}
	end = clock();
	elapsed_secs = double(end - begin) / CLOCKS_PER_SEC;
	cout << "Time elapsed: " << elapsed_secs << " seconds"<< endl;

	if (NLSR_flag==1) {
		/*
		 * calculate a spatial NNLS prior by non-local means
		 */
		float NL_control = atof(argv[3]);
		cout << endl;
		cout << "computing spatial priors..." << endl;
		begin = clock();
		calcSpatialNNLSprior(distributions, distributions_prior, rowN, columnN, sliceN, NL_control);
		end = clock();
		elapsed_secs = double(end - begin) / CLOCKS_PER_SEC;
		cout << "Time elapsed: " << elapsed_secs << " seconds"<< endl;

		/*
		 * starts srNNLS
		 */
		float reg_control = atof(argv[4]);
		cout << endl;
		cout << "calculating final T2 distributions..." << endl;
		begin = clock();
		for (int row=0 ; row<rowN ; row++) {
			// main loop
			for (int col=0 ; col<columnN ; col++) {
				for (int slice=0 ; slice<sliceN ; slice++) {
					// conditional loop to reject low signal pixels
					if ( T2_decay_map[slice*(rowN*columnN) + columnN*row + col] >= Threshold ) {
						for (int i=0 ; i<nechs ; i++) {
							decay_data[i] = (REAL)T2_decay_map[i*(sliceN*rowN*columnN) + slice*(rowN*columnN) + columnN*row + col];
						}
						/*
						 * Compute the NNLS basis over T2 space
						 */
						for (int x=0 ; x<nT2 ; x++) {
							EPGdecaycurveMKL2(nechs, alpha[IDX3C(row,col,slice,rowN,columnN)], TE, T2_times[x],
									T1, RefCon, echo_amp, Tr_dia, x, numCPU);
							for (int y=0; y<nechs ; y++) {
								basis_decay[IDX2C(x,y,nT2)] = (REAL)echo_amp[y];
							}
							priorT2[x] = (REAL)distributions_prior[IDX4C(row,col,slice,x,rowN,columnN,sliceN)];
						}
						/*
						 * Fit T2 distribution using NNLS (with regularization)
						 */
						srNNLS(basis_decay, decay_data, T2_dis_ls, Chi2Factor, priorT2, reg_control, numCPU);
						for (int x=0 ; x<nT2 ; x++) {
							distributions[IDX4C(row,col,slice,x,rowN,columnN,sliceN)] = (float)T2_dis_ls[x];
						}
					}
				}
			}
		}
		end = clock();
		elapsed_secs = double(end - begin) / CLOCKS_PER_SEC;
		cout << "Time elapsed: " << elapsed_secs << " seconds"<< endl;
	}

	/*
	 * Free GPU memory
	 */
	cudaFree(d_Am);
	cudaFree(d_Bm);
	cudaFree(d_Cm);

	/*
	 * generate MWF map
	 */
	// Verify the number of parameters in the command line
	cout << "\ngenerating MWF map..." << endl;
	float spwin1 = 0.0149;
	float spwin2 = 0.04;
	int sp[nT2];
	float sp_sum, dist_sum;
	for (int i=0;i<nT2;i++) {
		if ((T2_times[i] >= spwin1) && (T2_times[i] <= spwin2)) {
			sp[i] = 1;
		} else {
			sp[i] = 0;
		}
	}
	begin = clock();
	for (int row=0 ; row<rowN ; row++) {
		for (int col=0 ; col<columnN ; col++) {
			for (int slice=0 ; slice<sliceN ; slice++) {
				sp_sum = 0;
				dist_sum = 0;
				for (int x=0;x<nT2;x++) {
					dist_sum += distributions[IDX4C(row,col,slice,x,rowN,columnN,sliceN)];
					sp_sum += sp[x]*distributions[IDX4C(row,col,slice,x,rowN,columnN,sliceN)];
				}
				nii_data_float[slice*(rowN*columnN) + columnN*row + col] = sp_sum/dist_sum;
			}
		}
	}
	end = clock();
	elapsed_secs = double(end - begin) / CLOCKS_PER_SEC;
	cout << "Time elapsed: " << elapsed_secs << " seconds"<< endl;

	// write output file
	hdr.datatype = NIFTI_TYPE_FLOAT32;
	hdr.bitpix = 32;
	fp = fopen(argv[2],"w");
	if (fp == NULL) {
		fprintf(stderr, "\nError opening header file %s for write\n",argv[2]);
		exit(1);
	}
	/********** write first 348 bytes of header   */
	ret = fwrite(&hdr, MIN_HEADER_SIZE, 1, fp);
	if (ret != 1) {
		fprintf(stderr, "\nError writing header file %s\n",argv[2]);
		exit(1);
	}
	ret = fwrite(&pad, 4, 1, fp);
	if (ret != 1) {
		fprintf(stderr, "\nError writing header file extension pad %s\n",argv[2]);
		exit(1);
	}
	for (i=0; i<hdr.dim[1]*hdr.dim[2]*hdr.dim[3]*hdr.dim[4]; i++)
		nii_data_float[i] = (nii_data_float[i] - hdr.scl_inter)/hdr.scl_slope;
	ret = fwrite(nii_data_float, (size_t)(hdr.bitpix/8), hdr.dim[1]*hdr.dim[2]*hdr.dim[3]*hdr.dim[4], fp);
	if (ret != hdr.dim[1]*hdr.dim[2]*hdr.dim[3]*hdr.dim[4]) {
		fprintf(stderr, "\nError writing data to %s\n",argv[2]);
		exit(1);
	}
	fclose(fp);

	cudaPrintOff();

	free(nii_data);
	free(nii_data_float);
	delete [] T2_decay_map;
	delete [] priorT2;
	delete [] Tr_dia;
	delete [] decay_pred;
	delete [] concurrentT2_dis_ls;
	delete [] concurrentT2_dis_ls_single;
	delete [] T2_dis_ls;
	delete [] concurrentDecay_data;
	delete [] decay_data;
	delete [] alpha;
	delete [] distributions_prior;
	delete [] distributions;
	delete [] basis_decay;
	delete [] basis_angles;
	delete [] echo_amp;
	delete [] basis_angles_single;

	return EXIT_SUCCESS;
}

